from Crypto.Util import number
import math


#generates a n-bit prime number- fast for any length of n but advisable to have n less than 256
#requires: n to be at least greater than 8
def generate_prime(n):
    return number.getPrime(n)

#uses the unique id passed to extract a unique id for generating two primes
# returns a dict two primes
# requires : unique_id to be alphanumeric with a length greater than 10 and less than equal to 20
def hrsa_prime(unique_id):
    chars = 'qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM'
    nums ='0123456789'
    key = ''
    for char in unique_id:
        if char in nums:
            key+=char
        elif char in chars:
            key += str(chars.index(char))
    key = int(key)%256
    prime1 = generate_prime(key)
    prime2 = generate_prime(key)
    return {"prime1": prime1, "prime2": prime2}

# computes and returns a list of k primes starting from a prime
# which is "start" length of bits and ending with a prime which has
# a length of "end" bits
# requires: start>5
def generate_kprimes(k, start, end):
    numberofprimes = []
    if (k==end-start+1):
        bits = start
        for i in range(k):
            numberofprimes.append(number.getPrime(bits))
            bits+=1
        return numberofprimes
    else:
        return "K != end-start+1"


# computes the gcd of a and b, and returns the gcd
def gcd_euclid(a,b):
    return math.gcd(a,b)
#implements the extended euclidean algorithm to compute the coefficients
# and GCD of a and b of the linear diophantine equation found at the end
# of algorithm
def gcd_extended_euclid(a,b):

    if a == 0:
        return b,0,1
    gcd, x1, y1 = gcd_extended_euclid(b % a, a)
    # Update x and y using results of recursive
    # call
    x = y1 - (b // a) * x1
    y = x1

    return gcd, x, y






from .prime_euclid import *
import random

# computes and returns the public key of the two primes prime1 and prime2
# using RSA algorithm
def find_public(prime1, prime2):
    phi = (prime1-1)*(prime2-1)
    n = prime1*prime2
    public=0
    x = max(prime1, prime2)
    y = min(prime1, prime2)
    nonce = random.randint(y,x) # adds another layer of randomization
    for i in range(2, phi):
        if gcd_euclid(i,n)==1 and gcd_euclid(i,phi)==1:
            public = i
            if nonce==x or nonce ==y:
                break
    if public == 0:
        return "No key found"
    else:
        return public

# computes and returns the private key of the e,prime1,prime2 by finding the
# the inverse of e and phi(prime1*prime2) for which the gcd of e and phi is 1
def find_private(e,prime1,prime2):
    phi = (prime1-1)*(prime2-1)
    gcd,x,y = gcd_extended_euclid(e, phi)
    if gcd == 1:
        return x%phi
    else:
        return "Private key does not exist for the given values"

# generates a dict of the public key and the number to be used for modular operation
def generate_public_key(prime1,prime2):
    modulo = prime1*prime2
    res = find_public(prime1,prime2)

    if res == "No key found":
        raise Exception(res)
    else :
        return {"public": res, "modulo": modulo}

# generate a dict of the private key and the number to be used for modular operation
def generate_private_key(e,prime1,prime2):
    modulo = prime1*prime2
    res = find_private(e,prime1, prime2)

    if res == "Private key does not exist for the given values":
        raise Exception(res)
    else:
        return {"private": res, "modulo": modulo}




from .cryptodom import AES_encrypt
from .cryptodom import AES_decrypt
from .keys import generate_private_key
from .keys import generate_public_key
from .prime_euclid import generate_prime
from .prime_euclid import hrsa_prime
from .prime_euclid import generate_kprimes
from .pynacl import nacl_hashing
from .pynacl import get_password_hash
from .pynacl import verify_password_hash
from .pynacl import nacl_secret_key_encrypt_text
from .pynacl import nacl_secret_key_decrypt_text
from .rsacipher import hrsa_encrypted
from .rsacipher import hrsa_decrypted
from .vs import vigenere_decrypt
from .vs import vigenere_encrypt
from .vs import caesar_encrypt
from .vs import caesar_decrypt



from .keys import *

characters = ['null', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~', ' ', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', '_']

char_no_nums = ['null', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~', ' ', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', '_']



# encrypts the message using RSA and the index of the character stored in array characters
# devised an algorithm along with RSA to encrypt all characters with an ascii value greater than 32
# requires n to be greater than 100
def hrsa_encrypted(message,public,n):
    cipher = ''
    for char in message:
        pos = characters.index(char)
        encrypt = pow(pos, public, n)
        encrypt = str(encrypt)+char_no_nums[random.randint(2,60)]
        cipher+=encrypt
    return cipher

# decrypts the message using RSA and the index of the character stored in array characters
# devised an algorithm along with RSA to decrypt all characters with an ascii value greater than 32
# requires n to be greater than 100, which means p and q need to be chosen such that their product
# is greater than 100
def hrsa_decrypted(cipher,private,n):
    valid = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
    message = ''
    k=0
    length = len(cipher)
    while k<length:
        pos=''
        while k<length and cipher[k] in valid:
            pos+=cipher[k]
            k=k+1
        k+=1
        decrypt = pow(int(pos), private, n)
        message += characters[decrypt]
    return message









# --- vigenere cipher
# works best with lower case letters only

# implements the vigenere cipher on the given string, using the provided key
# the encrypted key is always given in lower case letters
def vigenere_encrypt(string, key):
    chars = 'abcdefghijklmnopqrstuvwxyz'
    string = string.lower()
    string_len = len(string)

    # checks and expands the key if its smaller than string
    key_len = len(key)
    while key_len < string_len:
        key +=key
        key_len = len(key)


    key_pos = 0
    encrypted =""
    for char in string:
        if char in chars:
            pos = chars.find(char)
            char_key = key[key_pos]
            char_key_pos = chars.find(char_key)
            key_pos += 1
            new_pos = (pos+char_key_pos)%26
            new_char = chars[new_pos]
            encrypted+=new_char
        else:
            encrypted+=char
    return encrypted



# implements the decryption of the vigenere cipher on the given string, using the provided key
def vigenere_decrypt(string, key):
    chars = 'abcdefghijklmnopqrstuvwxyz'
    string = string.lower()
    string_len = len(string)

    # checks and expands the key if its smaller than string
    key_len = len(key)
    while key_len < string_len:
        key += key
        key_len = len(key)

    key_pos = 0
    encrypted = ""
    for char in string:
        if char in chars:
            pos = chars.find(char)
            char_key = key[key_pos]
            char_key_pos = chars.find(char_key)
            key_pos += 1
            new_pos = (pos - char_key_pos) % 26
            new_char = chars[new_pos]
            encrypted += new_char
        else:
            encrypted += char
    return encrypted




#implements the caesar cipher and works on both upper case and lower case letters
# and  both negative positive shifts
def caesar_encrypt(word, shift):
    w = list(word)
    upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    lower = "abcdefghijklmnopqrstuvwxyz"
    for i in range(len(w)):
        if w[i] in upper:
            index = upper.index(w[i])
            f = index + shift % 26
            if f >= 26:
                f = abs(26 - (index + shift % 26))

            w[i] = upper[f]
        elif w[i] in lower:
            index = lower.index(w[i])
            f = index + shift % 26
            if f >= 26:
                f = abs(26 - (index + shift % 26))

            w[i] = lower[f]
    encrypted = ''.join(w)
    return encrypted


#implements the decryptiono of caesar cipher and works on both upper case and lower case letters
#and  both negative positive shifts
def caesar_decrypt(cipher, shift):
    return caesar_encrypt(cipher,-1*shift)

# string = "hamza yusuff is THE WORST hell ya 123"
# encrypt = caesar_encrypt(string,15)
# print(encrypt)
# text = caesar_decrypt(encrypt, 15)
# print(text)





from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

VERSION = '0.0.3'
PACKAGE_NAME = 'PyEnigmatic'
AUTHOR = 'Hamza Bin Yusuff'
AUTHOR_EMAIL = 'hbyusuff@uwaterloo.ca'
URL = 'https://github.com/hamza-yusuff'

LICENSE = 'MIT LICENSE'
DESCRIPTION = 'An enigmatic cryptographic library'
LONG_DESCRIPTION = long_description
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
      'cryptography',
      'pynacl',
      'pycryptodome'
]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      keywords=['python', 'enigma', 'cryptography', 'aes', 'rsa', 'caesar', 'vigenere', 'symmetric', 'assymetric', 'key', 'prime'],
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages()
      )



import nacl.secret
import nacl.utils
import nacl.encoding
import nacl.hash
import base64
import nacl.pwhash

# uses the sha 256 or sha 512 to create a hash for the given message
# and returns the message in string format
def nacl_hashing(message, sha=256):

    if sha == 512:
        hasher = nacl.hash.sha512
    else:
        hasher = nacl.hash.sha256
    message = message.encode('utf-8')
    digest = hasher(message, encoder=nacl.encoding.HexEncoder)
    return digest.decode('utf-8')


#password hashing uses key-streching algorithms like argon2, argon2i and scrypt
def get_password_hash(password, algo=None):
    if not isinstance(password, bytes):
        password = password.encode('utf-8')
    if not algo:
        hashed = nacl.pwhash.str(password)
    elif algo == "argon2i":
        hashed = nacl.pwhash.argon2i.str(password)
    elif algo == "scrypt":
        hashed = nacl.pwhash.scrypt.str(password)
    elif algo == 'argon2id':
        hashed = nacl.pwhash.argon2id.str(password)
    return base64.b64encode(hashed)

#verifies the given hash with the password
def verify_password_hash(password, hash):
    if not isinstance(password, bytes):
        password = password.encode('utf-8')
    hash = base64.b64decode(hash)
    try:
        nacl.pwhash.verify(hash, password)
        return True
    except:
        return False


# to create instances of secret key using nacl.secret.SecretBox
# for the encryption and decryption methods
# hashes the given key using sha256 to create the actual secret key for nacl.secret.SecretBox
class nacl_secretkey():

    def __init__(self,key,message):
        # initialises a secret key and message in byte format
        self.key = nacl_hashing(key)[:32].encode('utf-8')

        if isinstance(message, bytes):
            self.message= message
        else:
            self.message = message.encode('utf-8')
        if len(key) != 32:
            raise Exception("Not right length of 32 bytes")

        # uses the built in method to create an instance of the class of nacl.secret.SecretBox
        # that eventually encrypts and decrypts using XSalsa20 stream Cipher
        self.box = nacl.secret.SecretBox(self.key)

# encrypts the text using the given key
# returns the encrypted text (byte format)
# by performing a symmetric key encryption (XSalsa20 stream cipher)
def nacl_secret_key_encrypt_text(key,text):
    # encrypted_text = self.box.encrypt(self.message)
    if not isinstance(key, bytes):
        secret = nacl_secretkey(key, text)
    else:
        raise Exception("key is in bytes")
    encrypted_text = secret.box.encrypt(secret.message)
    # return encrypted_text,('\\').join(str(encrypted_text).split("\\"))[2:-1]]
    return base64.b64encode(encrypted_text)

# decrypts the text using the given key
# returns the decrypted text by performing a symmetric key encryption (XSalsa20 stream cipher)
def nacl_secret_key_decrypt_text(key, text):
    # plaintext = self.box.decrypt(self.encrypted)
    if not isinstance(key, bytes):
        secret = nacl_secretkey(key, text)
    else:
        raise Exception("Key or string is in bytes")

    plaintext = secret.box.decrypt(secret.message)
    return plaintext.decode('utf-8')


# text = nacl_secret_key_encrypt_text("qwertyuiopasdfghjklzxcvbnmqwerty","hamza")
# decrypted = nacl_secret_key_decrypt_text("qwertyuiopasdfghjklzxcvbnmqwerty", base64.b64decode(text))





from Crypto.Cipher import AES
from Crypto.Util.Padding import  pad
from Crypto.Util.Padding import unpad
import base64
#For AES key can be 128 or 192 or 256 bits

""""
aes_encrypt uses AES encryption to encrypt the given message
the mode of AES can be specified from either CBC or EAX
the function uses CBC as default
the key has to be either 128 or 192 or 256 bits
return a dictionary with initialization vector and ciphertext if CBC used and
return a dictionary with tag, nonce and ciphertext fields if EAX used """
def AES_encrypt(key, message, mode="CBC"):
    key_length = len(key)
    if (key_length==16 or key_length==24 or key_length==32):
        key = key.encode('utf-8')
        message = message.encode('utf-8')
        if (mode == "CBC"):
            cipher = AES.new(key, AES.MODE_CBC)

            initial_vector = base64.b64encode(cipher.iv)
            ciphertext = cipher.encrypt(pad(message, AES.block_size))
            ciphertext = base64.b64encode(ciphertext)

            return {"iv": initial_vector, "ciphertext": ciphertext}
        # elif (mode == "EAX"):
        #     cipher = AES.new(key, AES.MODE_EAX)
        #     ciphertext, tag = cipher.encrypt_and_digest(message)
        #     tag = base64.b64encode(tag)
        #     ciphertext = base64.b64encode(ciphertext)
        #     nonce = base64.b64encode(cipher.nonce)
        #     return {"tag":tag, "nonce":nonce, "ciphertext":ciphertext}
        else:
            raise Exception("Mode not supported")
    else:
        raise Exception("Key length has to be 128 or 192 or 256 bits")



"""
aes_decrypt uses AES cipher to decrypt the given encrypted message
the mode of AES can be specified from either CBC or EAX
the function uses CBC as default
the key has to be either 128 or 192 or 256 bits
parameter info should have tag,nonce and ciphertext fields if EAX mode used and
parameter info should have iv and ciphertext fields if CBC used
returns the plaintext or decrypted message """
def AES_decrypt(key, info,mode="CBC"):
    key_length = len(key)
    if (key_length == 16 or key_length == 24 or key_length == 32):

        if (mode=="CBC"):

            iv = base64.b64decode(info["iv"])
            ciphertext = base64.b64decode(info["ciphertext"])
            cipher_decrypt = AES.new(key.encode('utf-8'), AES.MODE_CBC, iv)
            plaintext = unpad(cipher_decrypt.decrypt(ciphertext), AES.block_size)
            return plaintext.decode()
        # elif (mode=="EAX"):
        #     nonce = base64.b64decode(info["nonce"])
        #     tag = base64.b64decode(info["tag"])
        #     ciphertext = base64.b64decode(info["ciphertext"])
        #     cipher = AES.new(key.encode('utf-8'), AES.MODE_EAX, nonce)
        #     data = cipher.decrypt_and_verify(ciphertext,tag)
        #     return data.decode()
        else:
            raise Exception("Mode not supported")

    else:
        raise Exception("Key length has to be 128 or 192 or 256 bits")


# data=aes_encrypt('mysecretpassword', 'this is my message',"CBC")
# print(aes_decrypt('mysecretpassword',data))